@@ -12,7 +12,6 @@
  */
 
 #include "c.h"
-#include "nls.h"
 #include "all-io.h"
 
 #include "blkdev.h"
@@ -42,7 +41,7 @@ struct fdisk_sgi_label {
 	} freelist[SGI_MAXPARTITIONS + 1];
 };
 
-static struct fdisk_parttype sgi_parttypes[] =
+static const struct fdisk_parttype sgi_parttypes[] =
 {
 	{SGI_TYPE_VOLHDR,	N_("SGI volhdr")},
 	{SGI_TYPE_TRKREPL,	N_("SGI trkrepl")},
@@ -124,10 +123,12 @@ static struct sgi_info *sgi_new_info(void)
 	info->b3 = cpu_to_be16(1);
 
 	/* You may want to replace this string !!!!!!! */
-	strcpy((char *) info->scsi_string, "IBM OEM 0662S12         3 30");
-	strcpy((char *) info->serial, "0000");
+	strncpy((char *) info->scsi_string, "IBM OEM 0662S12         3 30",
+		sizeof(info->scsi_string));
+	strncpy((char *) info->serial, "0000", sizeof(info->serial));
 	info->check1816 = cpu_to_be16(18 * 256 + 16);
-	strcpy((char *) info->installer, "Sfx version 5.3, Oct 18, 1994");
+	strncpy((char *) info->installer, "Sfx version 5.3, Oct 18, 1994",
+		sizeof(info->installer));
 
 	return info;
 }
@@ -154,7 +155,7 @@ int fdisk_sgi_create_info(struct fdisk_context *cxt)
 	/* I keep SGI's habit to write the sgilabel to the second block */
 	sgilabel->volume[0].block_num = cpu_to_be32(2);
 	sgilabel->volume[0].num_bytes = cpu_to_be32(sizeof(struct sgi_info));
-	strncpy((char *) sgilabel->volume[0].name, "sgilabel", 8);
+	memcpy((char *) sgilabel->volume[0].name, "sgilabel", 8);
 
 	fdisk_info(cxt, _("SGI info created on second sector."));
 	return 0;
@@ -300,7 +301,7 @@ static int sgi_get_disklabel_item(struct fdisk_context *cxt, struct fdisk_labeli
 		break;
 	default:
 		if (item->id < __FDISK_NLABELITEMS)
-			rc = 1;	/* unssupported generic item */
+			rc = 1;	/* unsupported generic item */
 		else
 			rc = 2;	/* out of range */
 		break;
@@ -397,20 +398,24 @@ static int sgi_check_bootfile(struct fdisk_context *cxt, const char *name)
 				   "e.g. \"/unix\" or \"/unix.save\"."));
 		return -EINVAL;
 
-	} else if (sz > sizeof(sgilabel->boot_file)) {
+	}
+
+	if (sz > sizeof(sgilabel->boot_file)) {
 		fdisk_warnx(cxt, P_("Name of bootfile is too long: %zu byte maximum.",
 				    "Name of bootfile is too long: %zu bytes maximum.",
 				    sizeof(sgilabel->boot_file)),
 			    sizeof(sgilabel->boot_file));
 		return -EINVAL;
 
-	} else if (*name != '/') {
+	}
+
+	if (*name != '/') {
 		fdisk_warnx(cxt, _("Bootfile must have a fully qualified pathname."));
 		return -EINVAL;
 	}
 
 	if (strncmp(name, (char *) sgilabel->boot_file,
-				sizeof(sgilabel->boot_file))) {
+				sizeof(sgilabel->boot_file)) != 0) {
 		fdisk_warnx(cxt, _("Be aware that the bootfile is not checked "
 				   "for existence.  SGI's default is \"/unix\", "
 				   "and for backup \"/unix.save\"."));
@@ -511,8 +516,8 @@ static int compare_start(struct fdisk_context *cxt,
 	 * Sort according to start sectors and prefer the largest partition:
 	 * entry zero is the entire-disk entry.
 	 */
-	unsigned int i = *(int *) x;
-	unsigned int j = *(int *) y;
+	const unsigned int i = *(const int *) x;
+	const unsigned int j = *(const int *) y;
 	unsigned int a = sgi_get_start_sector(cxt, i);
 	unsigned int b = sgi_get_start_sector(cxt, j);
 	unsigned int c = sgi_get_num_sectors(cxt, i);
@@ -592,9 +597,8 @@ static int verify_disklabel(struct fdisk_context *cxt, int verbose)
 		if (sgi_get_num_sectors(cxt, i) != 0) {
 			Index[sortcount++] = i;
 			if (sgi_get_sysid(cxt, i) == SGI_TYPE_ENTIRE_DISK
-			    && entire++ == 1) {
-				if (verbose)
-					fdisk_info(cxt, _("More than one entire "
+			    && entire++ == 1 && verbose) {
+				fdisk_info(cxt, _("More than one entire "
 						"disk entry present."));
 			}
 		}
@@ -691,7 +695,7 @@ static int verify_disklabel(struct fdisk_context *cxt, int verbose)
 		add_to_freelist(cxt, start, lastblock);
 	}
 	/*
-	 * Done with arithmetics. Go for details now.
+	 * Done with arithmetic. Go for details now.
 	 */
 	if (verbose) {
 		if (sgi_get_bootpartition(cxt) < 0
@@ -927,6 +931,7 @@ static int sgi_add_partition(struct fdisk_context *cxt,
 		fdisk_ask_number_set_default(ask, fdisk_scround(cxt, last) - 1);/* default */
 		fdisk_ask_number_set_high(ask,    fdisk_scround(cxt, last) - 1);/* maximal */
 		fdisk_ask_number_set_base(ask,    fdisk_scround(cxt, first));
+		fdisk_ask_number_set_wrap_negative(ask, 1); /* wrap negative around high */
 
 		if (fdisk_use_cylinders(cxt))
 			fdisk_ask_number_set_unit(ask,
@@ -984,9 +989,10 @@ static int sgi_create_disklabel(struct fdisk_context *cxt)
 			/* otherwise print error and use truncated version */
 			fdisk_warnx(cxt,
 				_("BLKGETSIZE ioctl failed on %s. "
-				  "Using geometry cylinder value of %llu. "
+				  "Using geometry cylinder value of %ju. "
 				  "This value may be truncated for devices "
-				  "> 33.8 GB."), cxt->dev_path, cxt->geom.cylinders);
+				  "> 33.8 GB."), cxt->dev_path,
+				(uintmax_t) cxt->geom.cylinders);
 		}
 	}
 
@@ -1005,7 +1011,8 @@ static int sgi_create_disklabel(struct fdisk_context *cxt)
 
 	/* sizeof(sgilabel->boot_file) = 16 > 6 */
 	memset(sgilabel->boot_file, 0, 16);
-	strcpy((char *) sgilabel->boot_file, "/unix");
+	strncpy((char *) sgilabel->boot_file, "/unix",
+		sizeof(sgilabel->boot_file));
 
 	sgilabel->devparam.skew			= (0);
 	sgilabel->devparam.gap1			= (0);
@@ -1180,13 +1187,11 @@ static const struct fdisk_label_operations sgi_operations =
 };
 
 /* Allocates an SGI label driver. */
-struct fdisk_label *fdisk_new_sgi_label(struct fdisk_context *cxt)
+struct fdisk_label *fdisk_new_sgi_label(struct fdisk_context *cxt __attribute__ ((__unused__)))
 {
 	struct fdisk_label *lb;
 	struct fdisk_sgi_label *sgi;
 
-	assert(cxt);
-
 	sgi = calloc(1, sizeof(*sgi));
 	if (!sgi)
 		return NULL;
@@ -1203,5 +1208,6 @@ struct fdisk_label *fdisk_new_sgi_label(struct fdisk_context *cxt)
 
 	lb->flags |= FDISK_LABEL_FL_REQUIRE_GEOMETRY;
 
-	return lb;
+	/* return calloc() result to keep static anaylizers happy */
+	return (struct fdisk_label *) sgi;
 }
